(*|  0:19  3/01/1998 *)
(* NoteOn/NoteOff - Lo() operator added *)
(*| 10:14 19/11/1997 *)
(* FindIOAddr - FOR loop replaced by REPEAT .. UNTIL *)
(*| 21:47 22/06/1996 *)
  Unit BWDrive;

  Interface

  Uses
    CRT;

  Type
    InstType = Array [0..15] Of Byte;

  Const
    SBDVersion = $0100;

  Var
    SBIOAddr : Word;

  Function FindIOAddr : Word;
{ Searches ports $220 - $260 for SoundBlaster card.  Returns 0 if not
  installed.  This routine sets SBIOAddr at startup time.

  NOTE:  To be compatible with AdLib cards, these FM music routines use
  ports $388 and $389.  If FindIOAddr returns 0, an AdLib card may still
  be installed.
}

  Function StatusRegister : Byte;
{ Returns the FM status register:

 Ŀ
     D7      D6      D5      D4      D3      D2      D1      D0    
 Ĵ
   IRQ         FLAG                                              
           T1     T2                                            
 
}

  Procedure InitCard;
{ This must be called to initialize the FM synthesizer chips. }

  Procedure SetTimer1(Count : Byte);
{ Timer1 is an 8 bit presettable counter.  It has a resolution of 80usec.
  Count is the preset value to be loaded into the counter.  If an overflow
  occurs, the Timer1 flag in the status register is set, and the preset
  value is reloaded into the counter.  In addition to normal timer func-
  tions, Timer1 is also used for control of composite speech synthesis. }

  Procedure SetTimer2(Count : Byte);
{ Timer2 is similar to Timer1 except that its resolution is 320usec. }

  Procedure TimerControl(Control : Byte);
{ This procedure is used to start, stop and control flags of the timers.

 Ŀ
     D7      D6      D5      D4      D3      D2      D1      D0    
 Ĵ
   IRQ         MASK                               START/STOP    
   RST     T1     T2                            T1      T2    
 

  D0 (Start/Stop Timer1)
    When this bit is 1, the preset value is loaded into Timer1 and
    counting starts.  When this bit is 0, Timer1 does not operate.
  D1 (Start/Stop Timer2)
    Performs the same function as D0 for Timer2.
  D5 (Mask Timer2)
    When this bit is 1, the flag for Timer2 in the status register is
    masked (always 0).
  D6 (Mask Timer1)
    Bit mask flag for Timer1.
  D7 (IRQ-RESET)
    When set to 1, the IRQ and Timer flags in the status register are
    reset to 0.  This bit then resets itself to 0. }

  Procedure CSMMode_KbdSplit(MKS : Byte);
{ This procedure causes the synthesizer to operate in music mode or speech
  synthesis mode and determines the keyboard split for keyboard scaling of
  rate.

 Ŀ
     D7      D6      D5      D4      D3      D2      D1      D0    
 Ĵ
   CSM     SEL                                                   
 

  D6 (NOTE SELECT)
    This bit controls the split point of the keyboard.  When 0, the key-
    board split is the second bit from the most significant bit of the
    F-Number.  The MSB of the F-Number is used when 1.

    NOTE SELECT = 0
 Ŀ
    OCTAVE      0      1      2      3      4      5      6      7   
 Ĵ
  BLOCK DATA    0      1      2      3      4      5      6      7   
 Ĵ
  F-NUMBER 9    1      1      1      1      1      1      1      1   
 Ĵ
  F-NUMBER 8  0  1  0  1  0  1  0  1  0  1  0  1  0  1  0  1 
 Ĵ
  SPLIT NUM   0  1  2  3  4  5  6  7  8  9 10 11 12 13 14 15 
 

    NOTE SELECT = 1
 Ŀ
    OCTAVE      0      1      2      3      4      5      6      7   
 Ĵ
  BLOCK DATA    0      1      2      3      4      5      6      7   
 Ĵ
  F-NUMBER 9  0  1  0  1  0  1  0  1  0  1  0  1  0  1  0  1 
 Ĵ
  SPLIT NUM   0  1  2  3  4  5  6  7  8  9 10 11 12 13 14 15 
 

  D7 (CSM)
    The composite sine wave speech synthesis mode is selected when 1.  All
    channels must be off when this mode is selected. }

  Procedure AM_VIB_EG_KSR_Multiple(InstNum, Modulator, Carrier : Byte);
{ This procedure controls several parameters affecting the timbre of the
  output signal.

 Ŀ
     D7      D6      D5      D4      D3      D2      D1      D0    
 Ĵ
    AM     VIB    EG     KSR              MULTIPLE             
 

  D0-D3 (MULTIPLE)
    The basic FM equation with the multiplication factor included is:

      F(t) = A sin (Mc Wc t + I sin (Mm Wm t))

    where A is the output amplitude
          I is the modulation index (amplitude of Modulator)
          Wc is the carrier frequency
          Wm is the modulator frequency
          Mc is the multiplication factor for Carrier cell
          Mm is the multiplication factor for Modulator cell

    The relationship between MULTIPLE and multiplication factor is:
    Ŀ  Ŀ
     MULTIPLE  MULTIPLICATION    MULTIPLE  MULTIPLICATION 
                   FACTOR                      FACTOR     
    Ĵ  Ĵ
         0           0.5             8            8       
         1            1              9            9       
         2            2              A           10       
         3            3              B           10       
         4            4              C           12       
         5            5              D           12       
         6            6              E           15       
         7            7              F           15       
      

  D4 (KEY SCALING RATE)
    When this bit is set to 1, the envelope length is gradually shortened
    as higher notes on the keyboard are played.
  D5 (ENVELOPE TYPE)
    A diminishing sound is selected when this bit is 0 and a continuous
    sound when it is 1.
  D6 (VIBRATO)
    Setting this bit to 1 cause a vibrato effect to be applied to the
    operator cell.
  D7 (AMPLITUDE MODULATION)
    Setting this bit to 1 causes tremolo to be applied to operator cell. }

  Procedure KSL_TotalLevel(InstNum, Modulator, Carrier : Byte);
{ This procedure controls the attenuation of the operator cells' output
  signal.

 Ŀ
     D7      D6      D5      D4      D3      D2      D1      D0    
 Ĵ
       KSL                           TOTAL LEVEL                  
 

  D0-D5 (TOTAL LEVEL)
    The TOTAL LEVEL attenuates the operator's output.  Varying the output
    level of an operator functioning as a carrier results in a change in
    the overall level of the sound contributed to the voice by that operator.
    This is how the volume changes.  Attenuating the output from a modulator
    will change the frequency spectrum produces by the carrier.  TOTAL LEVEL
    has a range of 0 to 63.  Maximum attenuation is 47.25 dB.  Degree of
    attenuation is given be the following formula.

      Attenuation = TOTAL LEVEL * 0.75 dB

  D6-D7 (KEYBOARD SCALING LEVEL)
    The KSL produces a gradual decrease in note output level towards higher
    pitch notes.  The degree of attenuation is controlled as follows:
    Ŀ
     D7  D6  Degree of Attenuation 
    Ĵ
      0   0            0           
      1   0        1.5dB/OCT       
      0   1         3dB/OCT        
      1   1         6dB/OCT        
    
}

  Procedure Attack_Decay(InstNum, Modulator, Carrier : Byte);
{ The ATTACK RATE is the rising time for the sound.  The DECAY RATE is the
  diminishing time after the attack.  These rates range from 0 to 15.

 Ŀ
     D7      D6      D5      D4      D3      D2      D1      D0    
 Ĵ
           ATTACK RATE                      DECAY RATE            
    2^3     2^2     2^1     2^0     2^3     2^2     2^1     2^0   
 
}

  Procedure Sustain_Release(InstNum, Modulator, Carrier : Byte);
{ For continuing sounds, the SUSTAIN LEVEL gives the point of change where
  the attenuated sound in the decay mode changes to a sound having a con-
  stant level.  For diminishing sounds, the sustain level gives the point
  where the decay mode changes to the release mode.

  For continuing sounds, the RELEASE RATE defines the rate at which the
  sound disappears after the notes is turned off.  For diminishing sounds,
  release rate indicates the attenuation after the sustain level is reached.

 Ŀ
     D7      D6      D5      D4      D3      D2      D1      D0    
 Ĵ
          SUSTAIN LEVEL                    RELEASE RATE           
    24dB    12dB    6dB     3dB     2^3     2^2     2^1     2^0   
 
}

  Procedure Block_FNum(InstNum, Modulator, Carrier : Byte);
{ This procedure determines the octave and frequency of the output signal.

 Ŀ
     D7      D6      D5      D4      D3      D2      D1      D0    
 Ĵ
                          F-NUMBER (L)                             
 Ĵ
                  KEYON          BLOCK           F-NUMBER (H)   
 

  D0-D7 (F-NUMBER L) & D0-D1 (F-NUMBER H)
    The F-Number is a 10 bit number that determines the frequency of the
    output signal.
  D2-D4 (BLOCK)
    These 3 buts determine the octave of the output signal.
  D5 (KEYON)
    When this be is 1, the channels is ON and voiced.  When 0, the channel
    is OFF.

  The relationship among F-NUMBER, BLOCK and output frequency is as follows:

    FREQUENCY = 50000 * F-NUMBER * 2^BLOCK-20

  FREQUENCY FOR BLOCK 4
  Ŀ
                              FNUM H            F-NUMBER L           
   NOTE  FREQUENCY  F-NUMBER Ĵ
                              D1  D0  D7  D6  D5  D4  D3  D2  D1  D0 
  Ĵ
    C#     277.2       363     0   1   0   1   1   0   1   0   1   1 
    D      293.7       385     0   1   1   0   0   0   0   0   0   1 
    D#     311.1       408     0   1   1   0   0   1   1   0   0   0 
    E      329.6       432     0   1   1   0   1   1   0   0   0   0 
    F      349.2       458     0   1   1   1   0   0   1   0   1   0 
    F#     370.0       485     0   1   1   1   1   0   0   1   0   1 
    G      392.0       514     1   0   0   0   0   0   0   0   1   0 
    G#     415.3       544     1   0   0   0   1   0   0   0   0   0 
    A      440.0       577     1   0   0   1   0   0   0   0   0   1 
    A#     466.2       611     1   0   0   1   1   0   0   0   1   1 
    B      493.9       647     1   0   1   0   0   0   0   1   1   1 
    C      523.3       686     1   0   1   0   1   0   1   1   1   0 
  
}

  Procedure AMVIBDepth_Rhythm(RegData : Byte);
{ This allows for control of amplitude modulation and vibrato depth, selec-
  tion of rhythm mode, and ON/OFF control of various rhythm instruments.

 Ŀ
     D7      D6      D5      D4      D3      D2      D1      D0    
 Ĵ
       DEPTH       RHY-  BASS   SNARE   TOM    TOP     HI    
    AM     VIB    THM   DRUM   DRUM    TOM    CYM     HAT   
 

  D0-D5 (RHYTHM)
    The synthesizer is set to rhythm sound mode when D5 is 1.  D0-D4 allow
    for ON/OFF control of the various rhythm instruments.  The last 3 FM
    channels must be off.  Slots 13-18 correspond to the rhythm sound shown
    below.  Data such as rate must be input as a value appropriate to each
    rhythm sound.
    Ŀ
     PERCUSSION INSTRUMENT  OPERATOR CELL 
    Ĵ
           Bass Drum             13,16    
           Hi Hat                14       
           Tom Tom               15       
           Snare Drum            17       
           Top Cymbal            18       
    
  D6 (VIB DEPTH)
    The vibrato depth is 14 cent when 1, 7 cent when 0.
  D7 (AM DEPTH)
    The tremolo depth is 4.8dB when 1, 1dB when 0.
}

  Procedure WaveSelect(InstNum, Modulator, Carrier : Byte);
{ When Wave Select is on, the output sine wave is distorted as shown below.

 Ŀ
     D7      D6      D5      D4      D3      D2      D1      D0    
 Ĵ
                                                     WAVE SEL     
 

    Ŀ
     D1  D0    Waveform   
    Ĵ
              Ŀ        
                      
      0   0  Ŀ  
                      
                      
              Ŀ        
      0   1           
               
              Ŀ  Ŀ  
      1   0       
              
              Ŀ   Ŀ   
      1   1         
               
    
}

  Procedure FeedBack_Connect(InstNum, RegData : Byte);
{ This procedure determines the modulation factor for self-feedback and
  the type of FM modulation.

 Ŀ
     D7      D6      D5      D4      D3      D2      D1      D0    
 Ĵ
                                        FEEDBACK           CON   
 

  D0 (CONNECTION)
    Connection controls the manner in which the two operator cells are
    connected.  The FM modulation mode is selected when this bit is 0.
    The two operator cells are connected in parallel for sine wave output
    in the composite mode when this bit is 1.
  D1-D3 (FEEDBACK)
    Gives the modulation factor for feedback FM modulation of the first
    operator cell.

Ŀ
  FEEDBACK     0      1      2      3      4      5      6      7   
Ĵ
 MODULATION                                                         
   FACTOR      0    II/16  II/8   II/4   II/2    II     2II    4II  

}

{ High-Level Procedures }

  Procedure LoadInstrument(Inst : Byte; InstData : InstType);
{ This procedure loads an instrument of InstType into the FM synthesizer.
  The instrument data starts at offset $24 of an SBI file.
}

  Procedure NoteOn(Inst, Oct, Note : Byte);
{ The procedure plays a note (1-12) in an octave (0-7) on channel (1-9).
  Note 1 is C#.
}

  Procedure NoteOff(Inst, Oct, Note : Byte);
{ The procedure stops a note (1-12) in an octave (0-7) on channel (1-9).
  (In this version, octave and note are ignored.)
}


Implementation

  Const
    ChanOffset : Array[1..9] OF Byte = (1,2,3,9,10,11,17,18,19);
    FNums : Array[0..12] Of Word = 
      ($157,$16B,$181,$198,$1B0,$1CA,$1E5,$202,$220,$241,$263,$287,$2AE);

  Function FindIOAddr : Word;
  Var
    Address : Word;
    Ready : Byte;
    Loop : Byte;
    Data : Byte;
  Begin
    FindIOAddr := 0;
    Address := $260;
    REPEAT
      Port[Address+6] := 1;
      Delay(1);
      Port[Address+6] := 0;
      Loop := 0;
      Repeat
        Ready := Port[Address+14];
        INC(Loop);
      Until ((Loop > 100) Or (Ready > $7F));
      If Ready > $7F Then Begin
        Data := Port[Address + 10];
        If Data = $AA Then Begin
          FindIOAddr := Address;
          Exit;
        End;
      End;
      Dec(Address,16);
    UNTIL (Address < $220)
  End;  {  }

  Function StatusRegister : Byte;
  Begin
    StatusRegister := Port[$388];
  End;  {  }

  Procedure InitCard;
  Begin
    Port[$388] := 1;
    Delay(1);
    Port[$389] := 0;
    Delay(1);
  End;  {  }

  Procedure SetTimer1(Count : Byte);
  Begin
    Port[$388] := 2;
    Delay(1);
    Port[$389] := Count;
    Delay(1);
  End;  {  }

  Procedure SetTimer2(Count : Byte);
  Begin
    Port[$388] := 3;
    Delay(1);
    Port[$389] := Count;
    Delay(1);
  End;  {  }

  Procedure TimerControl(Control : Byte);
  Begin
    Port[$388] := 4;
    Delay(1);
    Port[$389] := Control;
    Delay(1);
  End;  {  }

  Procedure CSMMode_KbdSplit(MKS : Byte);
  Begin
    Port[$388] := 8;
    Delay(1);
    Port[$389] := MKS;
    Delay(1);
  End;  {  }

  Procedure AM_VIB_EG_KSR_Multiple(InstNum, Modulator, Carrier : Byte);
  Begin
    Port[$388] := ChanOffset[InstNum] + $1F;
    Delay(1);
    Port[$389] := Modulator;
    Delay(1);
    Port[$388] := ChanOffset[InstNum] + $1F + 3;
    Delay(1);
    Port[$389] := Carrier;
    Delay(1);
  End;  {  }

  Procedure KSL_TotalLevel(InstNum, Modulator, Carrier : Byte);
  Begin
    Port[$388] := ChanOffset[InstNum] + $3F;
    Delay(1);
    Port[$389] := Modulator;
    Delay(1);
    Port[$388] := ChanOffset[InstNum] + $3F + 3;
    Delay(1);
    Port[$389] := Carrier;
    Delay(1);
  End;  {  }

  Procedure Attack_Decay(InstNum, Modulator, Carrier : Byte);
  Begin
    Port[$388] := ChanOffset[InstNum] + $5F;
    Delay(1);
    Port[$389] := Modulator;
    Delay(1);
    Port[$388] := ChanOffset[InstNum] + $5F + 3;
    Delay(1);
    Port[$389] := Carrier;
    Delay(1);
  End;  {  }

  Procedure Sustain_Release(InstNum, Modulator, Carrier : Byte);
  Begin
    Port[$388] := ChanOffset[InstNum] + $7F;
    Delay(1);
    Port[$389] := Modulator;
    Delay(1);
    Port[$388] := ChanOffset[InstNum] + $7F + 3;
    Delay(1);
    Port[$389] := Carrier;
    Delay(1);
  End;  {  }

  Procedure Block_FNum(InstNum, Modulator, Carrier : Byte);
  Begin
    Port[$388] := InstNum + $9F;
    Delay(1);
    Port[$389] := Modulator;
    Delay(1);
    Port[$388] := InstNum + $AF;
    Delay(1);
    Port[$389] := Carrier;
    Delay(1);
  End;  {  }

  Procedure AMVIBDepth_Rhythm(RegData : Byte);
  Begin
    Port[$388] := $BD;
    Delay(1);
    Port[$389] := RegData;
    Delay(1);
  End;  {  }

  Procedure WaveSelect(InstNum, Modulator, Carrier : Byte);
  Begin
    Port[$388] := ChanOffset[InstNum] + $DF;
    Delay(1);
    Port[$389] := Modulator;
    Delay(1);
    Port[$388] := ChanOffset[InstNum] + $DF + 3;
    Delay(1);
    Port[$389] := Carrier;
    Delay(1);
  End;  {  }

  Procedure FeedBack_Connect(InstNum, RegData : Byte);
  Begin
    Port[$388] := InstNum + $BF;
    Delay(1);
    Port[$389] := RegData;
    Delay(1);
  End;  {  }

  Procedure LoadInstrument(Inst : Byte; InstData : InstType);
  Begin
    AM_VIB_EG_KSR_Multiple(Inst, InstData[0], InstData[1]);
    KSL_TotalLevel(Inst, InstData[2], InstData[3]);
    Attack_Decay(Inst, InstData[4], InstData[5]);
    Sustain_Release(Inst, InstData[6], InstData[7]);
    WaveSelect(Inst, InstData[8], InstData[9]);
    FeedBack_Connect(Inst, InstData[10]);
  End;  {  }

  Procedure NoteOn(Inst, Oct, Note : Byte);
  Var
    B1,B2 : Byte;
    Note2 : Word;
  Begin
    If Oct > 7 Then
    Else Begin
      If Note > 12 Then
      Else Begin
        Note2 := FNums[Note];
        B2 := Hi(Note2) + 32;
        B2 := B2 + (Oct Shl 2);
        B1 := Lo(Note2);
        Block_FNum(Inst, B1, B2);
      End;
    End;
  End;  {  }

  Procedure NoteOff(Inst, Oct, Note : Byte);
  Var
    B1,B2 : Byte;
    Note2 : Word;
  Begin
    If Oct > 7 Then
    Else Begin
      If Note > 12 Then
      Else Begin
        Note2 := FNums[Note];
        B2 := Hi(Note2);
        B2 := B2 + (Oct Shl 2);
        B1 := Lo(Note2);
        Block_FNum(Inst, B1, B2);
      End;
    End;
  End;  {  }

Const
  Piano : InstType =
           ($21,$11,$4C,$00,$F1,$F2,$63,$72,$00,$00,$04,$00,$00,$00,$00,$00);
Var
  I : Byte;
Begin
  SBIOAddr := FindIOAddr;
(*
  IF SBIOAddr = 0 THEN BEGIN
    Writeln('No sound card detected, aborting');
    Halt(1);
  END;
*)
  InitCard;
  For I := 1 To 9 Do
    LoadInstrument(I, Piano);
End.

