(*| 22:35 19/11/1997 *)
PROGRAM FMSample;

{$DEFINE STEREO}
{$DEFINE OPL3}

(* -------------------------------------------------------------------------- */
/*                                                                            */
/* FM synthesizer low-level interface demo program.                           */
/* Copyright (c) 1994 Creative Labs, Inc.                                     */
/*                                                                            */
/* THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY      */
/* KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE        */
/* IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A PARTICULAR      */
/* PURPOSE.                                                                   */
/*                                                                            */
/* You have a royalty-free right to use, modify, reproduce and                */
/* distribute the Sample Files (and/or any modified version) in               */
/* any way you find useful, provided that you agree that                      */
/* Creative has no warranty obligations or liability for any Sample Files.    */
/*                                                                            */
/*----------------------------------------------------------------------------*/

/*
 *  This Program was originally written using Borland C++ Ver 3.1.
 *  Rewritten using Borland Pascal by Brian Whitnall
 *  To Compile : BPC FMSAMPLE
 * ---------------------------------------------------------------------
 *
 * This program is not intended to explain all the aspects of FM sound
 * generation on Sound Blaster cards.  The chips are too complicated for
 * that.  This program is just to demonstrate how to generate a tone and
 * control the left and right channels.  For more information on the FM
 * synthesizer chip, contact Yamaha.
 *
 * Here's a brief description of FM:  Each sound is created by two operator
 * cells (called "slots" in the Yamaha documentation), a modulator and a
 * carrier.  When FM synthesis was invented, the output value of the
 * modulator affected the frequency of the carrier.  In the Yamaha chips, the
 * modulator output actually affects the phase of the carrier instead of
 * frequency, but this has a similar  effect.
 *
 * Normally the modulator and carrier would probably be connected in series
 * for complex sounds.  For this program, I wanted a pure sine wave, so I
 * connected them in parallel and turned the modulator output down all the
 * way and used just the carrier.
 *
 * Sound Blaster 1.0 - 2.0 cards have one OPL-2 FM synthesis chip at
 * addresses 2x8 and 2x9 (base + 8 and base + 9).  Sound Blaster Pro version
 * 1 cards (CT-1330) achieve stereo FM with two OPL-2 chips, one for each
 * speaker.  The left channel FM chip is at addresses 2x0 and 2x1.  The right
 * is at 2x2 and 2x3.  Addresses 2x8 and 2x9 address both chips
 * simultaneously, thus maintaining compatibility with the monaural Sound
 * Blaster cards.  The OPL-2 contains 18 operator cells which make up the
 * nine 2-operator channels.  Since the CT-1330 SB Pro has two OPL-2 chips,
 * it is therefore capable of generating 9 voices in each speaker.
 *
 * Sound Blaster Pro version 2 (CT-1600) and Sound Blaster 16 cards have one
 * OPL-3 stereo FM chip at addresses 2x0 - 2x3.  The OPL-3 is separated into
 * two "banks."  Ports 2x0 and 2x1 control bank 0, while 2x2 and 2x3 control
 * bank 1.  Each bank can generate nine 2-operator voices.  However, when the
 * OPL-3 is reset, it switches into OPL-2 mode.  It must be put into OPL-3
 * mode to use the voices in bank 1 or the stereo features.  For stereo
 * control, each channel may be sent to the left, the right, or both
 * speakers, controlled by two bits in registers C0H - C8H.
 *
 * The FM chips are controlled through a set of registers.  The following
 * table shows how operator cells and channels are related, and the register
 * offsets to use.
 *
 * FUNCTION  MODULATOR-  -CARRIER--  MODULATOR-  -CARRIER--  MODULATOR-  -CARRIER--
 * OP CELL    1   2   3   4   5   6   7   8   9  10  11  12  13  14  15  16  17  18
 * CHANNEL    1   2   3   1   2   3   4   5   6   4   5   6   7   8   9   7   8   9
 * OFFSET    00  01  02  03  04  05  08  09  0A  0B  0C  0D  10  11  12  13  14  15
 *
 * An example will make the use of this table clearer:  suppose you want to
 * set the attenuation of both of the operators of channel 4.  The KSL/TOTAL LEVEL
 * registers (which set the attenuation) are 40H - 55H.  The modulator for
 * channel 4 is op cell 7, and the carrier for channel 4 is op cell 10.  The
 * offsets for the modulator and carrier cells are 08H and 0BH, respectively.
 * Therefore, to set the attenuation of the modulator, you would output a
 * value to register 40H + 08H == 48H, and to set the carrier's attenuation,
 * you would output to register 40H + 0BH == 4BH.
 *
 * In this program, I used just channel 1, so the registers I used were 20H,
 * 40H, 60H, etc., and 23H, 43H, 63H, etc.
 *
 * The frequencies of each channel are controlled with a frequency number and
 * a multiplier.  The modulator and carrier of a channel both get the same
 * frequency number, but they may be given different multipliers.  Frequency
 * numbers are programmed in registers A0H - A8H (low 8 bits) and B0H - B8H
 * (high 2 bits).  Those registers control entire channels (2 operators), not
 * individual operator cells.  To turn a note on, the key-on bit in the
 * appropriate channel register is set.  Since these registers deal with
 * channels, you do not use the offsets listed in the table above.  Instead,
 * add (channel-1) to A0H or B0H.  For example, to turn channel 1 on,
 * program the frequency number in registers A0H and B0H, and set the key-on
 * bit to 1 in register B0H.  For channel 3, use registers A2H and B2H.
 *
 * Bits 2 - 4 in registers B0H - B8H are the block (octave) number for the
 * channel.
 *
 * Multipliers for each operator cell are programmed through registers 20H -
 * 35H.  The table below shows what multiple number to program into the
 * register to get the desired multiplier.  The multiple number goes into
 * bits 0 - 3 in the register.  Note that it's a bit strange at the end.
 *
 *   multiple number     multiplier        multiple number     multiplier
 *          0                1/2                   8               8
 *          1                 1                    9               9
 *          2                 2                    10              10
 *          3                 3                    11              10
 *          4                 4                    12              12
 *          5                 5                    13              12
 *          6                 6                    14              15
 *          7                 7                    15              15
 *
 * This equation shows how to calculate the required frequency number (to
 * program into registers A0H - A8H and B0H - B8H) to get the desired
 * frequency:
 *                fn=(long)f * 1048576 / b / m /50000L
 * where f is the frequency in Hz,
 *       b is the block (octave) number between 0 and 7 inclusive, and
 *       m is the multiple number between 0 and 15 inclusive.
 *
 *)

USES Crt, Dos;

CONST
  KeyOn = $20;          { key-on bit in regs b0 - b8 }

 (* These are offsets from the base I/O address. *)
  FM = 8;               { SB (mono) ports (e.g. 228H and 229H) }
  ProFM1 = 0;           { On CT-1330, this is left OPL-2.  On CT-1600 and
                          later cards, it's OPL-3 bank 0. }
  ProFM2 = 2;           { On CT-1330, this is right OPL-2.  On CT-1600 and
                          later cards, it's OPL-3 bank 1. }
{$IFDEF OPL3}
  Left = $10;
  Right = $20;
{$ENDIF}

VAR
  IOPort: Word;         { Sound Blaster port address }

PROCEDURE MyDelay(Clocks: LongInt);
(*
 * "clocks" is clock pulses (at 1.193180 MHz) to elapse, but remember that
 * normally the system timer runs in mode 3, in which it counts down by twos,
 * so delay3(1193180) will only delay half a second.
 *
 *   clocks = time * 2386360
 *
 *     time = clocks / 2386360
 *)
VAR
  Elapsed: LongInt;
  Last, Next, NCopy, Diff, MSB: Word;
BEGIN
  Elapsed := 0;
  Port[$43] := 0;
  Last := Port[$40];
  MSB := Port[$40];
  Last := NOT((MSB SHL 8) + Last);
  REPEAT
    Port[$43] := 0;
    Next := Port[$40];
    MSB := Port[$40];
    Next := NOT((MSB SHL 8) + Next);
    NCopy := Next;
    Next := Next - Last;
    Elapsed := Elapsed + Next;
    Last := NCopy;
  UNTIL Elapsed >= Clocks
END;

FUNCTION Base16(VAR S: String): Word;
VAR
  S2: String;
  R, N: Integer;
BEGIN
  S2 := '$';
  WHILE S[1] <> ' ' DO BEGIN
    S2 := S2 + S[1];
    Delete(S, 1, 1);
  END;
  Val(S2, R, N);
  Base16 := R;
END;

FUNCTION Base10(VAR S: String): Word;
VAR
  S2: String;
  R, N: Integer;
BEGIN
  S2 := '';
  WHILE S[1] <> ' ' DO BEGIN
    S2 := S2 + S[1];
    Delete(S, 1, 1);
  END;
  Val(S2, R, N);
  Base10 := R;
END;

PROCEDURE ReadBlasterEnv(VAR PortN: Word; VAR Irq, Dma8, Dma16: Integer);
VAR
  Env: String;
  C: Char;
BEGIN
  Env := GetEnv('BLASTER');
  IF Env = '' THEN BEGIN
    Writeln('Unable to find BLASTER environment, aborting');
    Halt(1);
  END;
  WHILE Length(Env) > 0 DO BEGIN
    C := UpCase(Env[1]);
    Delete(Env, 1, 1);
    CASE C OF
      'A' : PortN := Base16(Env);
      'I' : Irq := Base10(Env);
      'D' : Dma8 := Base10(Env);
      'H' : Dma16 := Base10(Env);
    END;
  END;
END;

PROCEDURE FMoutput(PortN: Word; Reg, ValN: Integer);
(* This outputs a value to a specified FM register at a specified FM port. *)
BEGIN
  Port[PortN] := Reg;
  MyDelay(8);
  Port[PortN+1] := ValN;
  MyDelay(55);
END;

PROCEDURE FMP(Reg, ValN: Integer);
(* This function outputs a value to a specified FM register at the Sound
 * Blaster (mono) port address.
 *)
BEGIN
  FMoutput(IOPort + FM, Reg, ValN)
END;

PROCEDURE ProFM1P(Reg, ValN: Integer);
(* This function outputs a value to a specified FM register at the Sound
 * Blaster Pro left FM port address (or OPL-3 bank 0).
 *)
BEGIN
  FMoutput(IOPort + ProFM1, Reg, ValN)
END;

PROCEDURE ProFM2P(Reg, ValN: Integer);
(* This function outputs a value to a specified FM register at the Sound
 * Blaster Pro right FM port address (or OPL-3 bank 1).
 *)
BEGIN
  FMoutput(IOPort + ProFM2, Reg, ValN)
END;

PROCEDURE GetCh;
BEGIN
  REPEAT
  UNTIL ReadKey <> #0;
END;

PROCEDURE Main;
VAR
  I, Val1, Val2: Integer;
  Block, B, M, F, Fn: Word;
  FnL: LongInt;
  Dummy: Integer;
BEGIN
  ReadBlasterEnv(IOPort, Dummy, Dummy, Dummy);
{$IFDEF STEREO}
 {$IFDEF OPL3}
  Writeln('Program compiled for Sound Blaster Pro ver. 2 (CT-1600) and SB16 cards.');
 {$ELSE}
  Writeln('Program compiled for Sound Blaster Pro ver. 1 (CT-1330) cards.');
 {$ENDIF}
{$ELSE}
  Writeln('Program compiled for Sound Blaster 1.0 - 2.0 cards (monaural).');
{$ENDIF}
  FMP(1, 0);            { must initialize this to zero }
{$IFDEF OPL3}
  ProFM2P(5, 1);        { set to OPL3 mode, necessary for stereo }
  FMP($C0, LEFT + RIGHT + 1);  { set both channels, parallel connection }
{$ELSE}
  FMP($C0, 1);          { parallel connection }
{$ENDIF}

   (***************************************
    * Set parameters for the carrier cell *
    ***************************************)

  FMP($23, $21);   (* no amplitude modulation (D7=0), no vibrato (D6=0),
                    * sustained envelope type (D5=1), KSR=0 (D4=0),
                    * frequency multiplier=1 (D4-D0=1)
                    *)


  FMP($43, $00);   (* no volume decrease with pitch (D7-D6=0),
                    * no attenuation (D5-D0=0)
                    *)


  FMP($63, $FF);   { fast attack (D7-D4=0xF) and decay (D3-D0=0xF) }
  FMP($83, $05);   { high sustain level (D7-D4=0), slow release rate (D3-D0=5) }


   (*****************************************
    * Set parameters for the modulator cell *
    *****************************************)

  FMP($20, $20);   { sustained envelope type, frequency multiplier=0    }
  FMP($40, $3F);   { maximum attenuation, no volume decrease with pitch }

   (* Since the modulator signal is attenuated as much as possible, these
    * next two values shouldn't have any effect.
    *)
  FMP($60, $44);   { slow attack and decay }
  FMP($80, $05);   { high sustain level, slow release rate }


   {*************************************************
    * Generate tone from values looked up in table. *
    *************************************************}

  Writeln('440 Hz tone, values looked up in table.');
  FMP($A0, $41);        { 440 Hz }
  FMP($B0, $32);        { 440Hz, block 0, key on }
  GetCh;
  FMP($B0, $12);        { key off }


   (******************************************
    * Generate tone from a calculated value. *
    ******************************************)

  Writeln('440 Hz tone, values calculated.');
  Block := 4;
  M := 1;               { m is the frequency multple number }
  F := 440;             { want f=440 Hz }
  B := 1  SHL Block;

   { This is the equation to calculate frequency number from frequency. }

  FnL := F * 1048576 DIV B DIV M DIV 50000;
  Fn := Fnl;
  FMP($23, $20 OR (M AND $F));  (* 0x20 sets sustained envelope, low nibble
                                 * is multiple number
                                 *)
  FMP($A0, (Fn AND $FF));
  FMP($B0, (((Fn SHR 8) AND $03) + (Block SHL 2)) OR KeyOn);
  GetCh;


   (*********************************************************
    * Generate a range of octaves by changing block number. *
    *********************************************************)

  Writeln('Range of frequencies created by changing block number.');
  FOR Block := 0 TO 7 DO BEGIN
    Writeln('f=', 440*(1 SHL Block) DIV 16:5,' Hz (press Enter)');
    FMP($B0, (((Fn SHR 8) AND $03) + (Block SHL 2)) OR KeyOn);
    GetCh
  END;


   (*****************************************************************
    * Generate a range of frequencies by changing frequency number. *
    *****************************************************************)

   Writeln('Range of frequencies created by changing frequency number.');
   Block := 4;
   FOR Fn :=0 TO 1023 DO BEGIN
      FMP($A0, (Fn AND $FF));
      FMP($B0, (((Fn SHR 8) AND $03) + (Block SHL 2)) OR KEYON);
      Delay(1);
   END;

   (********************************************************************
    * Single tone again.  Both channels, then if on stereo board,      *
    * play tone in just the left channel, then just the right channel. *
    ********************************************************************)

  Writeln('440 Hz again, both channels.');
  Block := 4;
  Fn := 577;            { This number makes 440 Hz when block=4 and m=1 }
  FMP($A0, Fn AND $FF);
  FMP($B0, (((Fn SHR 8) AND $03) + (Block SHL 2)) OR KeyOn);
{$IFDEF STEREO}
 {$IFDEF OPL3}
    (* This left and right channel stuff is the only part of this program
     * that uses OPL3 mode.  Everything else is available on the OPL2.
     *)

  GetCh;
  Writeln('Left channel only');
  FMP($C0, LEFT OR 1);  { set left channel only, parallel connection }
  GetCh;
  Writeln('Right channel only');
  FMP($C0, RIGHT OR 1); { set right channel only, parallel connection }
 {$ELSE}
  GetCh;
  FMP($B0, ((Fn SHR 8) AND $03) + (Block SHL 2));       { key off }
  Writeln('Left channel only');
  ProFM1P($B0, (((Fn SHR 8) AND $03) + (Block SHL 2)) OR KeyOn);
  GetCh;
  ProFM1P($B0, ((Fn SHR 8) AND $03) + (Block SHL 2));   { key off }
  Writeln('Right channel only');
  ProFM2P($B0, (((Fn SHR 8) AND $03) + (Block SHL 2)) OR KeyOn);
 {$ENDIF}
{$ENDIF}


   (*********************************
    * Attenuate the signal by 3 dB. *
    *********************************)

  GetCh;
  FMP($B0, (((Fn SHR 8) AND $03) + (Block SHL 2)) OR KeyOn);
  Writeln('Attenuated by 3 db.');
  FMP($43, 4);          { attenuate by 3 db }
  GetCh;
  FMP($B0, ((Fn SHR 8) AND $03) + (Block SHL 2));
{$IFDEF OPL3}
   (* Set OPL-3 back to OPL-2 mode, because if the next program to run was
    * written for the OPL-2, then it won't set the LEFT and RIGHT bits to
    * one, so no sound will be heard.
    *)
  ProFM2P(5, 0);        { set back to OPL2 mode }
{$ENDIF}
END;

BEGIN
  Main;
END.
